/*
 *  GOL - Game of life simulation software
 *  Copyright (C) 1996 Michael Raymond Brown
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Author: Michael Raymond Brown
 *  EMail:  michael@arrakeen.demon.co.uk
 *  WWW:    http://www.arrakeen.demon.co.uk/
 *  Address:  Michael Brown
 *            Kia Ora
 *            Brook Street
 *            Great Bromley
 *            Nr Colchester
 *            Essex
 *            CO7 7HX
 *            England
 */


char*	    GOL_VERSION = "0.0.2";
const char* lasterrstr = "";
int	    lasterrnum = 0;

#include "App.h"
#include "Window.h"
#include "Button.h"
#include "CheckBox.h"
#include "Label.h"
#include "SpinBox.h"
#include "Scale.h"
#include "Text.h"

#include "mForm.h"
#include "mFrame.h"
#include "mRowCol.h"

#include "VScrollArea.h"
#include "ScrollArea.h"
#include "DrawArea.h"
#include "Image.h"
#include "Idle.h"

#include "MsgBox.h"

#include <iostream.h>

#include "life.h"
#include "png_life.h"
#include "lif_life.h"

#include <stdio.h>
#include <sys/times.h>
#include <sys/errno.h>
#include <errno.h>
#include <time.h>

#include "COPYING.h"

struct tms tms;






xwApp*			app;
class entry_form* entryform;

static char last_dir[256] = "";

class life_form : public xwHandler
   {public:
	life_form** preptr;
	life_form*    next;

	enum {m_open, m_save, m_exit
	    , m_clear, m_random};

	static xwMenu menu[];


	int    frame_count;
	int frame_preserve;
	clock_t  last_time;

	int	started_tick;
	Life*		life;
	xwWindow*      window;
	xwVScrollArea* scroll;

	xwWidget* toolarea;

	xwCheckBox*  pause;
	xwButton* singlestep;
	xwIdle*       idle;

	xwScale*     scaler;
	xwScale*    speeder;

	void enable_bar(int v)
	   {scaler->setEnabled(v);

	    window->menuSetEnable(m_clear, v);
	    window->menuSetEnable(m_random, v);
	    window->menuSetEnable(m_open, v);
	    window->menuSetEnable(m_save, v);
	   }

	xwImage*   image;

	int  width;
	int height;
	int scale_0[16][4];
	int scale_m1[256][2];
	char scale_m2[16];

	int scale;
	int delay;
	int white;
	int black;
	int greys[17];



	life_form(Life* l, life_form** list)
	   {
	    preptr = list;
	    if (next=*list) next->preptr=&this->next;
	    *list = this;

	    int count;
	    frame_count=0;
	    life=l;
	    started_tick=0;

	    width  = life->getWidth();
	    height = life->getHeight();

	    window = new xwWindow(app, "Life field", menu);
	    window->setHandler(this, -1);

	    xwWidget* toplevel = new xwmForm(window, 1);

	    scroll = new xwVScrollArea(toplevel , 0,0 , 1,1);
	    scroll->setRange(l->getWidth(), l->getHeight());
	    scroll->setSize(512, 400);
	    scroll->setHandler(this, 0);

	    image = app->getImage(width<512?width:512, height<400?height:400);

/************/
	    toolarea = new xwmRowCol(window, xwmRowCol::horizontal|xwmRowCol::pack_tight, 1);
	    window->addCommand(toolarea);

	    pause = new xwCheckBox(toolarea, "Pause");
	    pause->setHandler(this, 0);
	    pause->setValue(1);

	    singlestep = new xwButton(toolarea, "1 Cycle");
	    singlestep->setHandler(this, 0);

	    idle = new xwIdle(toolarea, app);
	    idle->setHandler(this, -2);

	    scaler = new xwScale(toolarea, "Zoom", -2, 2, xwScale::horizontal);
	    scaler->setValue(0); scale = 0;
	    scaler->setHandler(this, 0);

	    speeder = new xwScale(toolarea, "Speed", 1, 100, xwScale::horizontal);
	    speeder->setValue(100); delay = 0;
	    speeder->setHandler(this, 0);
/************/

	    window->show();

	    for (count=0; count<17; count++)
		 greys[count]=app->get_grey(255*count/16);

	    white = greys[16];
	    black = greys[ 0];

	    for (count=0; count<16; count++)
		{
		 scale_0[count][0]=(count&1)?white:black;
		 scale_0[count][1]=(count&2)?white:black;
		 scale_0[count][2]=(count&4)?white:black;
		 scale_0[count][3]=(count&8)?white:black;

		 scale_m2[count]=((count>>0)&1)+((count>>1)&1)
				+((count>>2)&1)+((count>>3)&1);
		}

	    for (count=0; count<256; count++)
		{int v1 = count&0x33;
		 int v2 = (count>>2)&0x33;

		 v1 = ((v1>>1)&0x11) + (v1&0x11);
		 v1 = ( v1>>4      ) + (v1&0x0f);
		 v2 = ((v2>>1)&0x11) + (v2&0x11);
		 v2 = ( v2>>4      ) + (v2&0x0f);

		 scale_m1[count][0]=greys[16*v1/4];
		 scale_m1[count][1]=greys[16*v2/4];
		}
	   }
	~life_form()
	   {*preptr = next;
	    if (next) next->preptr = preptr;
	    delete life;
	    if (image) delete image;
	   }
	void onDestroy(xwWindow* from, int index)
	   {
	    delete this;
	   }

	void update()
	   {int x, y, w, h;
	    scroll->getView(&x, &y, &w, &h);
	    onExpose(scroll, 0, 0, 0, w, h);
	   }

	void onClick(xwWidget* from, int index)
	   {
	    if (from == singlestep)
	       {
		frame_preserve=frame_count;
		last_time = times(&tms);
		enable_bar(0);
		if (pause->getValue()) idle->setDelay(0);
	       }
	   }
	void onChange(xwWidget* from, int index)
	   {if (from == speeder)
	       {delay = ((100-speeder->getValue())*1000)/99;
		if (pause->getValue()==0 && !started_tick)
		    idle->setDelay(delay);
	       }
	    else if (from == scaler)
	       {int x, y, w, h, nw, nh, p, q;
		int newscale = scaler->getValue();
		if (scale == newscale) return;
		scroll->getView(&x, &y, &w, &h);
		if (scale < 0)
		   {int s = 0-scale;
		    x=(x+w/2)<<s;
		    y=(y+h/2)<<s;
		   }
		else
		   {x=(x+w/2)>>scale;
		    y=(y+h/2)>>scale;
		   }
		if (newscale < 0)
		   {int s = 0-newscale;
		    x>>=s; y>>=s;
		    nw=width>>s;
		    nh=height>>s;
		   }
		else
		   {x<<=newscale; y<<=newscale;
		    nw=width<<newscale;
		    nh=height<<newscale;
		   }
		scroll->setOffsets(0, 0); // This stops motif warnings
		scroll->setRange(nw, nh);

		scroll->getView(&p, &q, &w, &h);
		x-=(w/2); y-=(h/2);
		if ((x+w) > nw) x=nw-w; if (x<0) x=0;
		if ((y+h) > nh) y=nh-h; if (y<0) y=0;
		scroll->setOffsets(x, y);

		scale = newscale;
		onResize(scroll, 0);
		scroll->clear();
	       }
	    else if (from == pause)
	       {if (pause->getValue() == 0)
		   {frame_preserve=frame_count;
		    last_time = times(&tms);
		    enable_bar(0);
		    idle->setDelay(delay);
		   }
		else if (!started_tick)
		   {
		    enable_bar(1);
		    idle->setDelay(-1);
	       }   }
	   }

	void onIdle(xwWidget* from, int index)
	   {started_tick=1;
	    if (life->tick())
	       {double fps;
		clock_t this_time;
		clock_t diff;

		started_tick=0;
		frame_count++;
		if (pause->getValue())
		   {enable_bar(1);
		    idle->setDelay(-1);
		   }
		else
		   {idle->setDelay(xwMsgBox::showing ? (delay>500 ? delay : 500) : delay);
		   }

		this_time = times(&tms);
		if ((diff=this_time-last_time) > 100)
		   {
		    fps=double(frame_count-frame_preserve) / double(diff) * 100;
		    fprintf(stderr, "\rFrames per second=%f", fps);
		    frame_preserve=frame_count;
		    last_time=this_time;
		   }

		update();
	       }
	    else
	       {idle->setDelay(0);
	   }   }


	void onResize(xwWidget* from, int index)
	   {if (from == scroll)
	       {int x, y, w, h;
		scroll->getView(&x,&y,&w,&h);
		if (w!=image->getWidth() || h!=image->getHeight())
		   {delete image;
		    image = app->getImage(w, h);
	   }   }   }


	void draw_scalem2(int x1, int y1, int x2, int y2)
	   {int x, dstx;
	    int y, dsty;
	    u_int bitmap[16];
	      int pixmap[4][4];
	      int *ptr;

	    scroll->getOffsets(&dstx, &dsty);
	    x1=(x1+dstx)*4; y1=(y1+dsty)*4;
	    x2=(x2+dstx)*4; y2=(y2+dsty)*4;

	    if (x1 < 0) x1=0;
	    if (y1 < 0) y1=0;
	    if (x2 >=  width) x2 =  width-1;
	    if (y2 >= height) y2 = height-1;
	    if (x1>x2 || y1>y2) return;

	    x1/=16; y1/=16;
	    x2/=16; y2/=16;

	    for (y=y1; y<=y2; y++) for (x=x1; x<=x2; x++)
		{life->getArea(x, y, bitmap);

#undef DOIT
#define DOIT(x, y)\
	pixmap[y/4][x/4]=greys[\
	 scale_m2[(bitmap[y+0]>>x)&15]+scale_m2[(bitmap[y+1]>>x)&15]\
	+scale_m2[(bitmap[y+2]>>x)&15]+scale_m2[(bitmap[y+3]>>x)&15]]
		 DOIT( 0, 0); DOIT( 4, 0); DOIT( 8, 0); DOIT( 12, 0);
		 DOIT( 0, 4); DOIT( 4, 4); DOIT( 8, 4); DOIT( 12, 4);
		 DOIT( 0, 8); DOIT( 4, 8); DOIT( 8, 8); DOIT( 12, 8);
		 DOIT( 0,12); DOIT( 4,12); DOIT( 8,12); DOIT( 12,12);

		 image->multiPlot((x*4)-dstx, (y*4)-dsty, 4, 4, (int*)pixmap, 4);
		}
	   }

	void draw_scalem1(int x1, int y1, int x2, int y2)
	   {int x, dstx;
	    int y, dsty;
	    u_int bitmap[16];
	      int pixmap[8][8];
	      int *ptr;

	    scroll->getOffsets(&dstx, &dsty);
	    x1=(x1+dstx)*2; y1=(y1+dsty)*2;
	    x2=(x2+dstx)*2; y2=(y2+dsty)*2;

	    if (x1 < 0) x1=0;
	    if (y1 < 0) y1=0;
	    if (x2 >=  width) x2 =  width-1;
	    if (y2 >= height) y2 = height-1;
	    if (x1>x2 || y1>y2) return;

	    x1/=16; y1/=16;
	    x2/=16; y2/=16;

	    for (y=y1; y<=y2; y++) for (x=x1; x<=x2; x++)
		{life->getArea(x, y, bitmap);

#undef DOIT
#define DOIT(y)\
	ptr=scale_m1[ ( bitmap[y]     &0xf)|((bitmap[y+1]<< 4)&0xf0) ];\
	pixmap[y/2][0]=ptr[0]; pixmap[y/2][1]=ptr[1];\
\
	ptr=scale_m1[ ((bitmap[y]>> 4)&0xf)|( bitmap[y+1]     &0xf0) ];\
	pixmap[y/2][2]=ptr[0]; pixmap[y/2][3]=ptr[1];\
\
	ptr=scale_m1[ ((bitmap[y]>> 8)&0xf)|((bitmap[y+1]>> 4)&0xf0) ];\
	pixmap[y/2][4]=ptr[0]; pixmap[y/2][5]=ptr[1];\
\
	ptr=scale_m1[ ((bitmap[y]>>12)&0xf)|((bitmap[y+1]>> 8)&0xf0) ];\
	pixmap[y/2][6]=ptr[0]; pixmap[y/2][7]=ptr[1]

		 DOIT( 0); DOIT( 2);
		 DOIT( 4); DOIT( 6);
		 DOIT( 8); DOIT(10);
		 DOIT(12); DOIT(14);

		 image->multiPlot((x*8)-dstx, (y*8)-dsty, 8, 8, (int*)pixmap, 8);
		}
	   }

	void draw_scale0(int x1, int y1, int x2, int y2)
	   {int x, dstx;
	    int y, dsty;
	    u_int bitmap[16];
	      int pixmap[16][16];
	      int *ptr;

	    scroll->getOffsets(&dstx, &dsty);
	    x1+=dstx; y1+=dsty;
	    x2+=dstx; y2+=dsty;

	    if (x1 < 0) x1=0;
	    if (y1 < 0) y1=0;
	    if (x2 >=  width) x2 =  width-1;
	    if (y2 >= height) y2 = height-1;
	    if (x1>x2 || y1>y2) return;

	    x1/=16; y1/=16;
	    x2/=16; y2/=16;

	    for (y=y1; y<=y2; y++) for (x=x1; x<=x2; x++)
		{life->getArea(x, y, bitmap);

#undef DOIT
#define DOIT(x, y)\
	ptr = scale_0[(bitmap[y]>>x) & 15];\
	pixmap[y][x+0]=ptr[0];\
	pixmap[y][x+1]=ptr[1];\
	pixmap[y][x+2]=ptr[2];\
	pixmap[y][x+3]=ptr[3]
#undef DOEM
#define DOEM(y)\
	DOIT( 0, y); DOIT( 4, y); DOIT( 8, y); DOIT(12, y)

		 DOEM( 0); DOEM( 1); DOEM( 2); DOEM( 3);
		 DOEM( 4); DOEM( 5); DOEM( 6); DOEM( 7);
		 DOEM( 8); DOEM( 9); DOEM(10); DOEM(11);
		 DOEM(12); DOEM(13); DOEM(14); DOEM(15);

		 image->multiPlot((x*16)-dstx, (y*16)-dsty, 16, 16, (int*)pixmap, 16);
		}
	   }

	void draw_scale1(int x1, int y1, int x2, int y2)
	   {int x, dstx;
	    int y, dsty;
	    u_int bitmap[16];
	      int pixmap[32][32];
	      int *ptr;

	    scroll->getOffsets(&dstx, &dsty);
	    x1=(x1+dstx)/2; y1=(y1+dsty)/2;
	    x2=(x2+dstx)/2; y2=(y2+dsty)/2;

	    if (x1 < 0) x1=0;
	    if (y1 < 0) y1=0;
	    if (x2 >=  width) x2 =  width-1;
	    if (y2 >= height) y2 = height-1;
	    if (x1>x2 || y1>y2) return;

	    x1/=16; y1/=16;
	    x2/=16; y2/=16;

	    for (y=y1; y<=y2; y++) for (x=x1; x<=x2; x++)
		{life->getArea(x, y, bitmap);

#undef DOIT
#define DOIT(x, y)\
	ptr = scale_0[(bitmap[y/2]>>(x/2)) & 15];\
	pixmap[y+0][x+0]=pixmap[y+0][x+1]=\
	pixmap[y+1][x+0]=pixmap[y+1][x+1]=ptr[0];\
	pixmap[y+0][x+2]=pixmap[y+0][x+3]=\
	pixmap[y+1][x+2]=pixmap[y+1][x+3]=ptr[1];\
	pixmap[y+0][x+4]=pixmap[y+0][x+5]=\
	pixmap[y+1][x+4]=pixmap[y+1][x+5]=ptr[2];\
	pixmap[y+0][x+6]=pixmap[y+0][x+7]=\
	pixmap[y+1][x+6]=pixmap[y+1][x+7]=ptr[3]
#undef DOEM
#define DOEM(y)\
	DOIT( 0, y); DOIT( 8, y); DOIT( 16, y); DOIT(24, y)

		 DOEM( 0); DOEM( 2); DOEM( 4); DOEM( 6);
		 DOEM( 8); DOEM(10); DOEM(12); DOEM(14);
		 DOEM(16); DOEM(18); DOEM(20); DOEM(22);
		 DOEM(24); DOEM(26); DOEM(28); DOEM(30);

		 image->multiPlot((x*32)-dstx, (y*32)-dsty, 32, 32, (int*)pixmap, 32);
		}
	   }

	void draw_scale2(int x1, int y1, int x2, int y2)
	   {int x, dstx;
	    int y, dsty;
	    u_int bitmap[16];
	      int pixmap[64][64];
	      int *ptr;

	    scroll->getOffsets(&dstx, &dsty);
	    x1=(x1+dstx)/4; y1=(y1+dsty)/4;
	    x2=(x2+dstx)/4; y2=(y2+dsty)/4;

	    if (x1 < 0) x1=0;
	    if (y1 < 0) y1=0;
	    if (x2 >=  width) x2 =  width-1;
	    if (y2 >= height) y2 = height-1;
	    if (x1>x2 || y1>y2) return;

	    x1/=16; y1/=16;
	    x2/=16; y2/=16;

	    for (y=y1; y<=y2; y++) for (x=x1; x<=x2; x++)
		{life->getArea(x, y, bitmap);

#undef DOIT
#define DOIT(x, y)\
	ptr = scale_0[(bitmap[y/4]>>(x/4)) & 15];\
	pixmap[y+0][x+ 0]=pixmap[y+0][x+ 1]=pixmap[y+0][x+ 2]=pixmap[y+0][x+ 3]=\
	pixmap[y+1][x+ 0]=pixmap[y+1][x+ 1]=pixmap[y+1][x+ 2]=pixmap[y+1][x+ 3]=\
	pixmap[y+2][x+ 0]=pixmap[y+2][x+ 1]=pixmap[y+2][x+ 2]=pixmap[y+2][x+ 3]=\
	pixmap[y+3][x+ 0]=pixmap[y+3][x+ 1]=pixmap[y+3][x+ 2]=pixmap[y+3][x+ 3]=ptr[0];\
	pixmap[y+0][x+ 4]=pixmap[y+0][x+ 5]=pixmap[y+0][x+ 6]=pixmap[y+0][x+ 7]=\
	pixmap[y+1][x+ 4]=pixmap[y+1][x+ 5]=pixmap[y+1][x+ 6]=pixmap[y+1][x+ 7]=\
	pixmap[y+2][x+ 4]=pixmap[y+2][x+ 5]=pixmap[y+2][x+ 6]=pixmap[y+2][x+ 7]=\
	pixmap[y+3][x+ 4]=pixmap[y+3][x+ 5]=pixmap[y+3][x+ 6]=pixmap[y+3][x+ 7]=ptr[1];\
	pixmap[y+0][x+ 8]=pixmap[y+0][x+ 9]=pixmap[y+0][x+10]=pixmap[y+0][x+11]=\
	pixmap[y+1][x+ 8]=pixmap[y+1][x+ 9]=pixmap[y+1][x+10]=pixmap[y+1][x+11]=\
	pixmap[y+2][x+ 8]=pixmap[y+2][x+ 9]=pixmap[y+2][x+10]=pixmap[y+2][x+11]=\
	pixmap[y+3][x+ 8]=pixmap[y+3][x+ 9]=pixmap[y+3][x+10]=pixmap[y+3][x+11]=ptr[2];\
	pixmap[y+0][x+12]=pixmap[y+0][x+13]=pixmap[y+0][x+14]=pixmap[y+0][x+15]=\
	pixmap[y+1][x+12]=pixmap[y+1][x+13]=pixmap[y+1][x+14]=pixmap[y+1][x+15]=\
	pixmap[y+2][x+12]=pixmap[y+2][x+13]=pixmap[y+2][x+14]=pixmap[y+2][x+15]=\
	pixmap[y+3][x+12]=pixmap[y+3][x+13]=pixmap[y+3][x+14]=pixmap[y+3][x+15]=ptr[3]
#undef DOEM
#define DOEM(y)\
	DOIT( 0, y); DOIT(16, y); DOIT(32, y); DOIT(48, y)

		 DOEM( 0); DOEM( 4); DOEM( 8); DOEM(12);
		 DOEM(16); DOEM(20); DOEM(24); DOEM(28);
		 DOEM(32); DOEM(36); DOEM(40); DOEM(44);
		 DOEM(48); DOEM(52); DOEM(56); DOEM(60);

		 image->multiPlot((x*64)-dstx, (y*64)-dsty, 64, 64, (int*)pixmap, 64);
		}
	   }



	void onExpose(xwWidget* from, int index, int x, int y, int w, int h)
	   {if (from == scroll)
	       {if (x<0) {w-=x; x=0; }
		if (y<0) {h-=y; y=0; }
		if ((x + w) >  image->getWidth()) w =  image->getWidth()-x;
		if ((y + h) > image->getHeight()) h = image->getHeight()-y;
		if (w<1 || h<1) return;
	        if (life->isOk()) switch (scale)
		   {
		    case -2:
			draw_scalem2(x, y, x+w-1, y+h-1);
			break;
		    case -1:
			draw_scalem1(x, y, x+w-1, y+h-1);
			break;
		    case 0:
			 draw_scale0(x, y, x+w-1, y+h-1);
			 break;
		    case 1:
			 draw_scale1(x, y, x+w-1, y+h-1);
			 break;
		    case 2:
			 draw_scale2(x, y, x+w-1, y+h-1);
			 break;
		   }
		image->putImage(x, y, x, y, w, h, scroll);
	   }   }


	void mouseDown(xwWidget* from, int index, int button, int x, int y)
	   {int p, q;
	    scroll->getOffsets(&p, &q);
	    switch (scale)
	       {
		case -2:
		    life->plot((x+p)*4, (y+q)*4, (button&button1)?1:0);
		    onExpose(scroll, index, x, y, 1, 1);
		    break;
		case -1:
		    life->plot((x+p)*2, (y+q)*2, (button&button1)?1:0);
		    onExpose(scroll, index, x, y, 1, 1);
		    break;
		case 0:
		    life->plot(x+p, y+q, (button&button1)?1:0);
		    onExpose(scroll, index, x, y, 1, 1);
		    break;
		case 1:
		    life->plot((x+p)/2, (y+q)/2, (button&button1)?1:0);
		    onExpose(scroll, index, ((x+p)&~1)-p, ((y+q)&~1)-q, 2, 2);
		    break;
		case 2:
		    life->plot((x+p)/4, (y+q)/4, (button&button1)?1:0);
		    onExpose(scroll, index, ((x+p)&~3)-p, ((y+q)&~3)-q, 4, 4);
		    break;
	       }
	   }
	void mouseMove(xwWidget* from, int index, int button, int x, int y)
	   {mouseDown(from, index, button, x, y);
	   }

	void onMenu(xwWindow* from, int index)
	   {
	    switch (index)
	       {
		case m_clear:
		    frame_count=0;
		    life->clear();
		    update();
		    break;
		case m_random:
		    frame_count=0;
		    life->randomize();
		    update();
		    break;

		case m_open:
		    {char result[256];
		     char  rules[22];
		     int len,w,h;
		     switch (xwMsgBox::file(app, "Open life field", result, last_dir, "*.png (Portable Network Graphic)|*.lif (Life format)"))
		        {case xwMsgBox::rsOK:
			     cerr << "ok with " << result << "\n";
			     life->clear();
			     len=strlen(result);
			     if ((len>4) && !strcmp(result+len-4, ".png"))
				{if (open_png_life(life, result, &w, &h, rules) == -1)
				    {sprintf(result, "Error %d (%s) while opening\n", lasterrnum, lasterrstr);
				     xwMsgBox::ask(app, "Open life field", result);
				 } else update();
				}
			     else if ((len>4) && !strcmp(result+len-4, ".lif"))
				{if (open_lif_life(life, result, &w, &h, rules) == -1)
				    {sprintf(result, "Error %d (%s) while opening\n", lasterrnum, lasterrstr);
				     xwMsgBox::ask(app, "Open life field", result);
				 } else update();
				}
			     else
				{xwMsgBox::ask(app, "Open field", "unrecognised file extension");
				}
			     break;
			 case xwMsgBox::rsCANCEL:
			     cerr << "cancel\n";
			     break;
			 case xwMsgBox::rsUNKNOWN:
			     cerr << "unknown\n";
			     break;
			 default:
			     cerr << "default\n";
		        }
		    }
		    break;
		case m_save:
		    {char result[256];
		     int     len;
		     switch (xwMsgBox::file(app, "Save field as", result, last_dir, "*.png (Portable Network Graphic)"))
		        {case xwMsgBox::rsOK:
			     cerr << "ok with " << result << "\n";
			     len=strlen(result);
			     if ((len>4) && !strcmp(result+len-4, ".png"))
				{if (save_png_life(life, result) == -1)
				    {sprintf(result, "Error %d (%s) while saving\n", lasterrnum, lasterrstr);
				     xwMsgBox::ask(app, "Save field as", result);
				}   }
			     else if ((len>4) && !strcmp(result+len-4, ".lif"))
				{if (save_lif_life(life, result) == -1)
				    {sprintf(result, "Error %d (%s) while saving\n", lasterrnum, lasterrstr);
				     xwMsgBox::ask(app, "Save field as", result);
				}   }
			     else
				{xwMsgBox::ask(app, "Save field as", "unrecognised file extension");
				}
			     break;
			 case xwMsgBox::rsCANCEL:
			     cerr << "cancel\n";
			     break;
			 case xwMsgBox::rsUNKNOWN:
			     cerr << "unknown\n";
			     break;
			 default:
			     cerr << "default\n";
		        }
		    }
		    break;
		case m_exit:
		    delete window;
		    break;
	       }
	   }
   };
xwMenu life_form::menu[] =
	   { "File", 'F', -1, xwMenu::submenu
	       , "Open"   , 'O', m_open, xwMenu::item
	       , "Save As", 'S', m_save, xwMenu::item
	       , 0        , 0  ,      0, xwMenu::seperator
	       , "Exit"   , 'E', m_exit, xwMenu::item
	       , 0        , 0  , 0     , xwMenu::END
	   , "Edit", 'E', -1, xwMenu::submenu
	       , "Clear" , 'C', m_clear , xwMenu::item
	       , "Random", 'R', m_random, xwMenu::item
	       , 0       , 0  , 0       , xwMenu::END
	   , 0, 0, 0, xwMenu::END};



class entry_form : public xwHandler
   {public:
	enum {m_open, m_exit};
	static xwMenu menu[];

	xwWindow*  window;
	xwmForm*     form;
	xwButton*   pushb[10];
	xwCheckBox*toggle[9];
	int	 bitcount;

	xwButton*  quit_b;
	xwButton*    go_b;

	xwSpinBox*  width;
        xwSpinBox* height;
	xwCheckBox*  wrap;

	life_form* formlist;

	entry_form()
	   {
	    formlist = (life_form*)0;

	    xwWidget*   rules;
	    xwWidget* bitmask;
	    xwWidget* buttons;
	    xwWidget*    size;
	    xwmFrame*   frame;

	    window = new xwWindow(app, "Game_of_life", menu);
	    window->setHandler(this, -5);
	    form = new xwmForm(window, 100);

		frame = new xwmFrame(form, 0, 0, 70, 70);
	    rules   = new xwmForm(frame, 10);

		frame = new xwmFrame(form, 70, 0, 100, 40);
	    bitmask = new xwmForm(frame, 3);

		frame = new xwmFrame(form, 70, 40, 100, 70);
	    buttons = new xwmForm(frame, 2);

		frame = new xwmFrame(form, 0, 70, 100, 100);
	    size    = new xwmForm(frame, 100);


	    (pushb[0] = new xwButton(rules, "0: Clear", 0, 0, 5, 2))->setHandler(this, Life::CLEAR);
	    (pushb[1] = new xwButton(rules, "1: Clear", 0, 2, 5, 4))->setHandler(this, Life::CLEAR);
	    (pushb[2] = new xwButton(rules, "2: Leave", 0, 4, 5, 6))->setHandler(this, Life::LEAVE);
	    (pushb[3] = new xwButton(rules, "3: Set"  , 0, 6, 5, 8))->setHandler(this, Life::SET  );
	    (pushb[4] = new xwButton(rules, "4: Clear", 0, 8, 5,10))->setHandler(this, Life::CLEAR);
	    (pushb[5] = new xwButton(rules, "5: Clear", 5, 0,10, 2))->setHandler(this, Life::CLEAR);
	    (pushb[6] = new xwButton(rules, "6: Clear", 5, 2,10, 4))->setHandler(this, Life::CLEAR);
	    (pushb[7] = new xwButton(rules, "7: Clear", 5, 4,10, 6))->setHandler(this, Life::CLEAR);
	    (pushb[8] = new xwButton(rules, "8: Clear", 5, 6,10, 8))->setHandler(this, Life::CLEAR);
	    (pushb[9] = new xwButton(rules, "9: Clear", 5, 8,10,10))->setHandler(this, Life::CLEAR);
	    pushb[9]->setEnabled(0);
	    bitcount=8;

	    (toggle[0] = new xwCheckBox(bitmask, "", 0, 0, 1, 1))->setValue(1); toggle[0]->setHandler(this, 0);
	    (toggle[1] = new xwCheckBox(bitmask, "", 1, 0, 2, 1))->setValue(1); toggle[1]->setHandler(this, 1);
	    (toggle[2] = new xwCheckBox(bitmask, "", 2, 0, 3, 1))->setValue(1); toggle[2]->setHandler(this, 2);
	    (toggle[3] = new xwCheckBox(bitmask, "", 0, 1, 1, 2))->setValue(1); toggle[3]->setHandler(this, 3);
	    (toggle[4] = new xwCheckBox(bitmask, "", 1, 1, 2, 2))->setValue(0); toggle[4]->setHandler(this, 4);
	    (toggle[5] = new xwCheckBox(bitmask, "", 2, 1, 3, 2))->setValue(1); toggle[5]->setHandler(this, 5);
	    (toggle[6] = new xwCheckBox(bitmask, "", 0, 2, 1, 3))->setValue(1); toggle[6]->setHandler(this, 6);
	    (toggle[7] = new xwCheckBox(bitmask, "", 1, 2, 2, 3))->setValue(1); toggle[7]->setHandler(this, 7);
	    (toggle[8] = new xwCheckBox(bitmask, "", 2, 2, 3, 3))->setValue(1); toggle[8]->setHandler(this, 8);

	    (go_b   = new xwButton(buttons, "GO!" , 0, 0, 2, 1))->setHandler(this, -1);
	    (quit_b = new xwButton(buttons, "Quit", 0, 1, 2, 2))->setHandler(this, -2);

	    new xwLabel(size, "Width: ", 0, 0, 20, 50);
	    new xwLabel(size, "Height: ", 0, 50, 20, 100);
	    (width  = new xwSpinBox(size, 64, 32768, 64, 20, 0, 50, 50))->setValue(256);
	    (height = new xwSpinBox(size, 64, 32768, 64, 20, 50, 50, 100))->setValue(256);
	    (wrap = new xwCheckBox(size, "Wrap arround at edges", 50, 0, 100, 50))->setValue(1);

	    window->show();
	   }

	Life* go()
	   {char bitmask[3][3];
	    char   rules[10];
	    int    count;

	    for (count=9; count--; )
		 bitmask[count%3][count/3]=toggle[count]->getValue();
	    for (count=10; count--; )
		 rules[count]=pushb[count]->getIndex();

	    return new local_Life(width->getValue(), height->getValue(), wrap->getValue(), rules, bitmask);
	   }

	void onClick(xwWidget* from, int value)
	   {char caption[16];
	    switch (value)
		   {case Life::CLEAR:
			 from->getCaption(caption);
			 strcpy(caption+3, "Leave");
			 from->setCaption(caption);
			 from->setIndex(Life::LEAVE);
			 break;
		    case Life::LEAVE:
			 from->getCaption(caption);
			 strcpy(caption+3, "Set");
			 from->setCaption(caption);
			 from->setIndex(Life::SET);
			 break;
		    case Life::SET:
			 from->getCaption(caption);
			 strcpy(caption+3, "Clear");
			 from->setCaption(caption);
			 from->setIndex(Life::CLEAR);
			 break;

		    case -1: // The 'GO' button

			 new life_form(go(), &formlist);
			 break;
		    case -2: // The 'quit' button
			 if (formlist && xwMsgBox::ask(app, "Quit", "There are still active simulations.\nAre you sure?")!=xwMsgBox::rsYES)
			     break;
			 delete app;
			 exit(0);
			 break;
		   }
	   }
	void onChange(xwWidget* from, int index)
	   {if (from->getValue())
	       {bitcount++;
		pushb[bitcount]->setEnabled(1);
	       }
	    else
	       {pushb[bitcount]->setEnabled(0);
		bitcount--;
	       }
	   }

	void set_button(int n, int val)
	   {const char* cap;
	    char tmp[64];
	    if (pushb[n]->getIndex()==val) return;
	    switch (val)
	       {case Life::CLEAR:
		    cap="Clear";
		    break;
		case Life::LEAVE:
		    cap="Leave";
		    break;
		case Life::SET:
		    cap="Set";
		    break;
	       }
	    pushb[n]->getCaption(tmp);
	    strcpy(tmp+3, cap);
	    pushb[n]->setCaption(tmp);
	    pushb[n]->setIndex(val);
	   }
	void set_toggle(int n, int val)
	   {int v = toggle[n]->getValue();
	    if ((v&&!val) || (val&&!v))
	       {toggle[n]->setValue(val);
		onChange(toggle[v], 0);
	       }
	   }
	void set_wrap(int val)
	   {int v = wrap->getValue();
	    if ((v&&!val) || (val&&!v)) wrap->setValue(val);
	   }

	void onMenu(xwWindow* from, int index)
	   {
	    switch (index)
	       {
		case m_open:
		    {char result[256];
		     char  rules[22];
		     int len,w,h;
		     if (xwMsgBox::file(app, "Open life field", result, last_dir, "*.png (Portable Network Graphic)") == xwMsgBox::rsOK)
			{
			 cerr << "ok with " << result << "\n";
			 len=strlen(result);
			 if ((len>4) && !strcmp(result+len-4, ".png"))
			    {
			     if (open_png_life((Life*)0, result, &w, &h, rules) == -1)
				{sprintf(result, "Error %d (%s) while opening\n", lasterrnum, lasterrstr);
				 xwMsgBox::ask(app, "Save field as", result);
				}
			     w=(w+63)&~63;
			     h=(h+63)&~63;
			     if (rules[0])
				{char* ptr;
				 set_wrap(rules[0] == 'y');
				 ptr=rules+2;
				 for (len=0; len<9; len++)
				      set_toggle(len, ptr[len]=='x');
				 ptr=rules+12;
				 for (len=0; len<10; len++)
				      set_button(len, (ptr[len]=='c')?int(Life::CLEAR):((ptr[len]=='l')?int(Life::LEAVE):int(Life::SET)));
				}
			     else if (xwMsgBox::ask(app, "Open life field", "The file does not have a GOL header, it will \nuse the default settings from the setup window. \nContinue?")!=xwMsgBox::rsYES)
				{break;
				}
			     width->setValue(w);
			     height->setValue(h);
			     Life* life = go();
			     if (open_png_life(life, result, &w, &h, rules) == -1)
				{delete life;
				 sprintf(result, "Error %d (%s) while opening\n", lasterrnum, lasterrstr);
				 xwMsgBox::ask(app, "Save field as", result);
				}
			     new life_form(life, &formlist);
			    }
			 else
			    {xwMsgBox::ask(app, "Open life field", "unrecognised file extension");
			    }
			}
		    }
		    break;
		case m_exit:
		    if (formlist && xwMsgBox::ask(app, "Quit", "There are still active simulations.\nAre you sure?")!=xwMsgBox::rsYES)
			break;
		    delete app;
		    exit(0);
		    break;
	       }
	   }
	void onDestroy(xwWindow* from, int index)
	   {delete app;
	    exit(0);
	   }
   };
xwMenu entry_form::menu[] =
	   { "File", 'F', -1, xwMenu::submenu
	       , "Open"   , 'O', m_open, xwMenu::item
	       , 0        , 0  ,      0, xwMenu::seperator
	       , "Exit"   , 'E', m_exit, xwMenu::item
	       , 0        , 0  , 0     , xwMenu::END
	   , 0, 0, 0, xwMenu::END};





class copyright : public xwHandler
   {public:
	xwWindow	*window;
	xwmForm		*form;
	xwScrollArea	*scroll;
	xwText		*text;
	xwButton	  *ok;

	copyright()
	   {window = new xwWindow(app, "Game of life");
	    window->setHandler(this, 0);

	    form = new xwmForm(window, 50);
	    scroll = new xwScrollArea(form, 0, 0, 50, 46);
	    text = new xwText(scroll, 80, COPYRIGHT_LINES);

	    text->setEditable(0);
	    text->setText(COPYRIGHT_TEXT);

	    (ok = new xwButton(form, "Continue",  22, 46,  28, 50))->setHandler(this, 0);

	    window->show();
	    window->center();
	   }

	void onClick(xwWidget* from, int index)
	   {if (from == ok) {
		entryform = new entry_form();
		delete window;
	   }}

	void onDestroy(xwWindow* from, int index)
	   {delete this;
	   }
   };





class iownthis : public xwHandler
   {public:
	xwWindow	*window;
	xwmForm		*form;
	xwText		*text;
	xwButton	*cont;
	xwButton	*view;
	int	   makeitsafe;

	iownthis()
	   {window = new xwWindow(app, "Game of life");
	    window->setHandler(this, 0);

	    form = new xwmForm(window, 14);
	    text = new xwText(form, 64, 6, 0, 0, 14, 11);

	    text->setEditable(0);
	    char buffer[512];
	    sprintf(buffer,
"GOL release %s, Copyright (C) 1996 Michael Raymond Brown\n"
"GOL comes with ABSOLUTELY NO WARRANTY.\n"
"This is free software, and you are welcome to redistribute it\n"
"under certain conditions; Click `View GPL\' for details.\n"
"GOL can be found at http://www.arrakeen.demon.co.uk/", GOL_VERSION);
	    text->setText(buffer);

	    (cont = new xwButton(form, "Continue",  3, 11,  5, 14))->setHandler(this, 0);
	    (view = new xwButton(form, "View GPL",  9, 11, 11, 14))->setHandler(this, 0);

	    makeitsafe=0;
	    window->show();
	    window->center();
	   }

	void onClick(xwWidget* from, int index)
	   {if (from == cont) {
		makeitsafe=1;
		delete window;
		entryform = new entry_form();
	    } else if (from == view) {
		makeitsafe=1;
		delete window;
		new copyright();
	   }}

	void onDestroy(xwWindow* from, int index)
	   {if (!makeitsafe)
	       {delete app;
		exit(0);
	       }
	    delete this;
	   }
   };





int main(int argc, char** argv)
   {
    app = new xwApp("Game_of_life", argc, argv);

    new iownthis();

    app->start();
    return 0;
   }

